@@ -31,6 +31,7 @@ import java.util.Set;
 import java.util.stream.Stream;
 import javax.annotation.Nonnull;
 import org.apache.flink.api.common.ExecutionConfig;
+import org.apache.flink.api.common.JobID;
 import org.apache.flink.api.common.accumulators.Accumulator;
 import org.apache.flink.api.common.accumulators.DoubleCounter;
 import org.apache.flink.api.common.accumulators.Histogram;
@@ -99,12 +100,13 @@ public class ReductionsTest {
             new FakeKeyedStateBackend(),
             new FakeTimerServiceFactory(),
             new FakeInternalListState(),
+            new FakeMapState<>(),
             new HashMap<>(),
             new FakeOutput(),
             TestUtils.ENVELOPE_FACTORY,
             MoreExecutors.directExecutor(),
             new FakeMetricGroup(),
-            new FakeMapState());
+            new FakeMapState<>());
 
     assertThat(reductions, notNullValue());
   }
@@ -302,6 +304,11 @@ public class ReductionsTest {
     public void registerUserCodeClassLoaderReleaseHookIfAbsent(String s, Runnable runnable) {
       throw new UnsupportedOperationException();
     }
+
+    @Override
+    public JobID getJobId() {
+      throw new UnsupportedOperationException();
+    }
   }
 
   private static final class FakeKeyedStateBackend implements KeyedStateBackend<Object> {
@@ -352,9 +359,9 @@ public class ReductionsTest {
 
     @Nonnull
     @Override
-    public <T extends HeapPriorityQueueElement & PriorityComparable & Keyed>
+    public <T extends HeapPriorityQueueElement & PriorityComparable<? super T> & Keyed<?>>
         KeyGroupedInternalPriorityQueue<T> create(
-            @Nonnull String stateName, @Nonnull TypeSerializer<T> byteOrderedElementSerializer) {
+            @Nonnull String s, @Nonnull TypeSerializer<T> typeSerializer) {
       throw new UnsupportedOperationException();
     }
 
@@ -511,44 +518,44 @@ public class ReductionsTest {
     }
   }
 
-  private static final class FakeMapState implements MapState<Long, Message> {
+  private static final class FakeMapState<K, V> implements MapState<K, V> {
 
     @Override
-    public Message get(Long key) throws Exception {
+    public V get(K key) throws Exception {
       return null;
     }
 
     @Override
-    public void put(Long key, Message value) throws Exception {}
+    public void put(K key, V value) throws Exception {}
 
     @Override
-    public void putAll(Map<Long, Message> map) throws Exception {}
+    public void putAll(Map<K, V> map) throws Exception {}
 
     @Override
-    public void remove(Long key) throws Exception {}
+    public void remove(K key) throws Exception {}
 
     @Override
-    public boolean contains(Long key) throws Exception {
+    public boolean contains(K key) throws Exception {
       return false;
     }
 
     @Override
-    public Iterable<Entry<Long, Message>> entries() throws Exception {
+    public Iterable<Entry<K, V>> entries() throws Exception {
       return null;
     }
 
     @Override
-    public Iterable<Long> keys() throws Exception {
+    public Iterable<K> keys() throws Exception {
       return null;
     }
 
     @Override
-    public Iterable<Message> values() throws Exception {
+    public Iterable<V> values() throws Exception {
       return null;
     }
 
     @Override
-    public Iterator<Entry<Long, Message>> iterator() throws Exception {
+    public Iterator<Entry<K, V>> iterator() throws Exception {
       return null;
     }
 
