@@ -21,14 +21,22 @@ package org.apache.flink.statefun.flink.datastream;
 import java.net.URI;
 import java.time.Duration;
 import org.apache.flink.annotation.Internal;
+import org.apache.flink.shaded.jackson2.com.fasterxml.jackson.databind.ObjectMapper;
+import org.apache.flink.shaded.jackson2.com.fasterxml.jackson.databind.node.ObjectNode;
+import org.apache.flink.statefun.flink.core.httpfn.DefaultHttpRequestReplyClientSpec;
 import org.apache.flink.statefun.flink.core.httpfn.HttpFunctionEndpointSpec;
-import org.apache.flink.statefun.flink.core.jsonmodule.FunctionEndpointSpec.Target;
-import org.apache.flink.statefun.flink.core.jsonmodule.FunctionEndpointSpec.UrlPathTemplate;
+import org.apache.flink.statefun.flink.core.httpfn.TargetFunctions;
+import org.apache.flink.statefun.flink.core.httpfn.TransportClientConstants;
+import org.apache.flink.statefun.flink.core.httpfn.TransportClientSpec;
+import org.apache.flink.statefun.flink.core.httpfn.UrlPathTemplate;
 import org.apache.flink.statefun.sdk.FunctionType;
 
 /** A Builder for RequestReply remote function type. */
 public class RequestReplyFunctionBuilder {
 
+  private final DefaultHttpRequestReplyClientSpec.Timeouts transportClientTimeoutsSpec =
+      new DefaultHttpRequestReplyClientSpec.Timeouts();
+
   /**
    * Create a new builder for a remote function with a given type and an endpoint.
    *
@@ -46,7 +54,8 @@ public class RequestReplyFunctionBuilder {
   private RequestReplyFunctionBuilder(FunctionType functionType, URI endpoint) {
     this.builder =
         HttpFunctionEndpointSpec.builder(
-            Target.functionType(functionType), new UrlPathTemplate(endpoint.toASCIIString()));
+            TargetFunctions.functionType(functionType),
+            new UrlPathTemplate(endpoint.toASCIIString()));
   }
 
   /**
@@ -57,7 +66,7 @@ public class RequestReplyFunctionBuilder {
    * @return this builder.
    */
   public RequestReplyFunctionBuilder withMaxRequestDuration(Duration duration) {
-    builder.withMaxRequestDuration(duration);
+    transportClientTimeoutsSpec.setCallTimeout(duration);
     return this;
   }
 
@@ -68,7 +77,7 @@ public class RequestReplyFunctionBuilder {
    * @return this builder.
    */
   public RequestReplyFunctionBuilder withConnectTimeout(Duration duration) {
-    builder.withConnectTimeoutDuration(duration);
+    transportClientTimeoutsSpec.setConnectTimeout(duration);
     return this;
   }
 
@@ -79,7 +88,7 @@ public class RequestReplyFunctionBuilder {
    * @return this builder.
    */
   public RequestReplyFunctionBuilder withReadTimeout(Duration duration) {
-    builder.withReadTimeoutDuration(duration);
+    transportClientTimeoutsSpec.setReadTimeout(duration);
     return this;
   }
 
@@ -90,7 +99,7 @@ public class RequestReplyFunctionBuilder {
    * @return this builder.
    */
   public RequestReplyFunctionBuilder withWriteTimeout(Duration duration) {
-    builder.withWriteTimeoutDuration(duration);
+    transportClientTimeoutsSpec.setWriteTimeout(duration);
     return this;
   }
 
@@ -107,6 +116,20 @@ public class RequestReplyFunctionBuilder {
 
   @Internal
   HttpFunctionEndpointSpec spec() {
+    final TransportClientSpec transportClientSpec =
+        new TransportClientSpec(
+            TransportClientConstants.OKHTTP_CLIENT_FACTORY_TYPE,
+            transportClientPropertiesAsObjectNode(transportClientTimeoutsSpec));
+    builder.withTransport(transportClientSpec);
     return builder.build();
   }
+
+  private static ObjectNode transportClientPropertiesAsObjectNode(
+      DefaultHttpRequestReplyClientSpec.Timeouts transportClientTimeoutsSpec) {
+    final DefaultHttpRequestReplyClientSpec transportClientSpecPojo =
+        new DefaultHttpRequestReplyClientSpec();
+    transportClientSpecPojo.setTimeouts(transportClientTimeoutsSpec);
+
+    return new ObjectMapper().valueToTree(transportClientSpecPojo);
+  }
 }
