@@ -18,12 +18,11 @@
 
 package org.apache.flink.statefun.flink.datastream;
 
-import java.util.Collections;
-import java.util.Map;
 import java.util.Objects;
 import javax.annotation.Nullable;
 import javax.annotation.concurrent.NotThreadSafe;
 import org.apache.flink.annotation.Internal;
+import org.apache.flink.statefun.flink.core.httpfn.DefaultHttpRequestReplyClientFactory;
 import org.apache.flink.statefun.flink.core.httpfn.HttpFunctionEndpointSpec;
 import org.apache.flink.statefun.flink.core.httpfn.HttpFunctionProvider;
 import org.apache.flink.statefun.sdk.FunctionType;
@@ -35,17 +34,17 @@ final class SerializableHttpFunctionProvider implements SerializableStatefulFunc
 
   private static final long serialVersionUID = 1;
 
-  private final Map<FunctionType, HttpFunctionEndpointSpec> supportedTypes;
+  private final HttpFunctionEndpointSpec spec;
   private transient @Nullable HttpFunctionProvider delegate;
 
-  SerializableHttpFunctionProvider(Map<FunctionType, HttpFunctionEndpointSpec> supportedTypes) {
-    this.supportedTypes = Objects.requireNonNull(supportedTypes);
+  SerializableHttpFunctionProvider(HttpFunctionEndpointSpec spec) {
+    this.spec = Objects.requireNonNull(spec);
   }
 
   @Override
   public StatefulFunction functionOfType(FunctionType type) {
     if (delegate == null) {
-      delegate = new HttpFunctionProvider(supportedTypes, Collections.emptyMap());
+      delegate = new HttpFunctionProvider(spec, DefaultHttpRequestReplyClientFactory.INSTANCE);
     }
     return delegate.functionOfType(type);
   }
